#!/usr/bin/perl -I/usr/sausalito/perl
#
# $Id: vz_vserver_backup.pl, v 1.1.0.13 Thu 02 Feb 2012 05:56:17 PM CET mstauber Exp $
# Copyright 2006-2012 Stauber Multimedia Design. All rights reserved.
#
# This script performs local and remote backing up of virtual private servers. 
# Only VPS's which have either local or remote backup enabled in the GUI will be backed up!
#

use CCE;
use POSIX qw(isalpha);
use Socket;
use Sys::Hostname;
use File::Spec;
use List::MoreUtils qw(uniq);
use POSIX qw(strftime);
use MIME::Lite;
use Net::SFTP::Foreign;
use File::Basename;

my $cce = new CCE;
$cce->connectuds();

$oid = $cce->event_oid();
$obj = $cce->event_object();

# Helper to redirect SSH errors to /dev/null:
open $ssherr, '>', '/dev/null'; # or die "Unable to open /dev/null";

$host = hostname();
$addr = inet_ntoa(scalar(gethostbyname($host)) || 'localhost');
$now = localtime time;
$datetime = strftime('%Y-%m-%d %H:%M:%S', localtime);

$DEBUG = "0";

# Get local / remote switch:
$type_switch = $ARGV[0];
if ($type_switch eq "local") {
	$b_type = "local";
	$B_type = "Local";
	$s_type = "local";
}
elsif ($type_switch eq "remote") {
	$b_type = "remote";
	$B_type = "Remote";
	$s_type = "remote";
}
else {
	$b_type = "local & remote";
	$B_type = "Local & Remote";
	$s_type = "both";
}

# Find out if there is anything to back up (we may be cluster secondary after all!)
if (!-d "/vz/private") {
    print STDERR "INFO: This node is currently secondary in a cluster.\n";
    print STDERR "So there is nothing to backup on this node.\n\n";
    exit 1;
}

# Get settings out of CODB to determine which type of backup is enabled:
@oids = $cce->find('VserverMasterNode');
if (!defined($oids[0])) {
        print STDERR "VZBACKUP: Sorry, no VserverMasterNode object in CCE found!\n";
        exit 1;
}

# Remove stale lockfiles if they're older than 180 minutes:  
if (-f "/tmp/vz_vserver_backup.lock") {
    system("find /tmp/vz_vserver_backup.lock -type f -cmin +180 -print | xargs rm >/dev/null 2>&1");
}

# Create lockfile:
if (-f "/tmp/vz_vserver_backup.lock") {
    print STDERR "ERROR: Found lockfile /tmp/vz_vserver_backup.lock!\n";
    print STDERR "Please check that no other instance of this program runs!\n\n";
    exit 1;
}

($ok, $solbackupValues) = $cce->get($oids[0]);
if ($ok) {
    if (($solbackupValues->{'vzdump_available'} == "1") && ($solbackupValues->{'vzdump_enabled'} == "1")) {
	$using_vzdump = "1";
	# Use VZdump:
	&type_vzdump;
    }
    elsif ($solbackupValues->{'rsync_enabled'} == "1") {
	$using_vzdump = "0";
	# Use RSYNC:
	&type_rsync;
    }
    else {
	# Do nothing:
        print STDERR "INFO: Backup is not configured on this node.\n";
        print STDERR "So there is nothing to do here. Please configure\n";
        print STDERR "the backup through the GUI first!\n";
    }
    # Remove lockfile:
    system("/bin/rm -f /tmp/vz_vserver_backup.lock");
}

$cce->bye('SUCCESS');
exit 0;

sub backup_kvm_mailhead {
    if ($using_vzdump == "1") {
	# Email generation:
	open (EMAIL, '>>/tmp/.backup_mail');
	print EMAIL "=================================================================\n";
	print EMAIL "Aventurin{e} $B_type KVM Backup on $host ($addr) \n";
	print EMAIL "Starting up at: $now\n";
        print EMAIL "=================================================================\n\n";
    }
}

sub backup_kvm_mailfoot {
    # Email separately if we use VZdump:
    if ($using_vzdump == "1") {
	$now = localtime time;
        print EMAIL "=================================================================\n";
        print EMAIL "Aventurin{e} $B_type KVM Backup on $host ($addr) \n";
        print EMAIL "Backup finished at: $now\n";
        print EMAIL "=================================================================\n\n";
        close (EMAIL);

        # Send the email:
        open (EMAIL, '/tmp/.backup_mail');
	    while (<EMAIL>) {
		$body .= $_;
	    }
	close (EMAIL);
        system("/bin/rm -f /tmp/.backup_mail");

	if ($body) {
	    $body = $body_head . $body;
	    $subject = $host . ": $B_type KVM backup finished";

	    # Build the message using MIME::Lite instead:
	    my $send_msg = MIME::Lite->new(
    		From     => "root <root>",
    		To       => "root <root>",
    		Subject  => $subject,
    		Data     => $body
	    );   
        
	    # Set content type:
	    $send_msg->attr("content-type"         => "text/plain");
	    $send_msg->attr("content-type.charset" => "ISO-8859-1"); 
     
	    # Out with the email:
	    $send_msg->send;
	}
    }
}

## Local Backup KVM's:
sub backup_kvm_local {

    # Get settings for RSYNC related backup:
    &rsync_settings;

    # Handle local RSYNC backup if local backup is enabled and supposed to run now:
    if (($MN_LocalBackupEnabled == "1") && (($s_type eq "local") || ($s_type eq "both"))) {
	print STDERR "Starting local backup of KVM's ... \n\n";
	print EMAIL "Starting local backup of KVM's ... \n";
	print EMAIL "=========================================\n\n";

	# Find all KVM's where local backup is enabled:
	$nolocal = "0";
	@oids = $cce->find('Vserver_Backup', {'VserverBackupLocalEnabled' => "1", 'virttype' => 'kvm'});
	if ($#oids == -1) {
    	    $nolocal = "1";
	}
	else {
	    # Get all KVM's with enabled local backup:
	    foreach $line (@oids) {
    		($ok, $system) = $cce->get($line);
    		$func = push(@vps_local, $system->{vpsid});
	    }
	    @sorted_includes = sort { $a <=> $b } @vps_local;
	}

	# Handle early exit if nothing to do:
	if ($nolocal == "1") {
	    print STDERR "No KVM has local backup enabled. \n";
	    print EMAIL "No KVM has local backup enabled. \n\n";
	}
	else {

	    # Start the backup of KVM VPS's:
	    foreach $vpsid (@sorted_includes) {
		$local_target_dir_vps = $local_location_rsync . "/" . $vpsid;
        	($sys_oid) = $cce->find('KVM', {'vpsname' => $vpsid});
        	($ok, $sysvserver) = $cce->get($sys_oid);
		$vps_fqdn = $sysvserver->{'hostname'};
		$vps_ip = $sysvserver->{'ip'};
		print STDERR "Performing local backup of KVM $vpsid to $local_target_dir_vps\n";
		print EMAIL "Performing local backup of KVM $vpsid to $local_target_dir_vps\n";

		# Get KVM settings:
		@kvmoid = $cce->find('KVM', {'vpsname' => $vpsid});
		($ok, $kvm) = $cce->get($kvmoid[0]);
		$kvmdisk = $kvm->{'disk'};

		if (!-d $local_target_dir_vps) {
		    system("mkdir -p $local_target_dir_vps");
		}

		# Get just the name of the filename of the disk image:
		$base_name = basename($kvmdisk);

		# We delete the old image first to save space. Otherwise we double our disk usage during the time
		# when the image is backed up:
		system("rm -f $local_location_rsync/$vpsid/$base_name");

		# Do the backup:
		system("/usr/bin/rsync -q -ar $kvmdisk $local_location_rsync/$vpsid/$base_name --delete-after");
		system("cp /etc/libvirt/qemu/$vpsid.xml $local_location_rsync/$vpsid/");

		# As local backup has just ended, we now update CODB with the info that says so:
        	@vps_oid = $cce->find('Backup', {'vpsid' => $vpsid, 'origin' => 'rsync', 'location' => 'local', 'virttype' => 'kvm'});
        	if ($#vps_oid < 0) {
		    # No CODB object for this type of backup yet. Creating one:
		    ($ok) = $cce->create('Backup', {
    			'origin' => 'rsync',
    			'vpsid' => $vpsid,
    			'virttype' => 'kvm',
    			'location' => 'local',
    			'LBtimestamp' => $datetime,
    			'storage' => $local_target_dir_vps,
    			'ip' => '',
    			'hostname' => '',
    			'timestamp' => time()
		    });
        	}
		else {
		    # Object already present - updating it:
		    ($ok) = $cce->set($vps_oid[0], '', {
    			'vpsid' => $vpsid,
    			'LBtimestamp' => $datetime,
    			'storage' => $local_target_dir_vps,
    			'ip' => '',
    			'hostname' => '',
    			'timestamp' => time()
		    });
		}
		print STDERR "Finished local backup of KVM $vpsid \n\n";
		print EMAIL "Finished local backup of KVM $vpsid \n\n";
	    }
	}
    }
}

## Remote Backup KVM's:
sub backup_kvm_remote {

    # Get settings for RSYNC related backup:
    &rsync_settings;

    # Handle remote RSYNC backup if remote backup is enabled and supposed to run now:
    if (($MN_RemoteBackupEnabled == "1") && (($s_type eq "remote") || ($s_type eq "both"))) {
	print STDERR "Starting remote backup of KVM's ... \n\n";
	print EMAIL "Starting remote backup of KVM's ... \n";
	print EMAIL "=========================================\n\n";

	# Find all KVM VPS's where remote backup is enabled:
	@oids = $cce->find('Vserver_Backup', {'VserverBackupRemoteEnabled' => "1", 'virttype' => 'kvm'});
	$noremote = "0";
	if ($#oids == -1) {
    	    $noremote = "1";
	}
	else {
	    # Get all VPS's with enabled remote backup:
	    foreach $line (@oids) {
    		($ok, $system) = $cce->get($line);
    		$func = push(@vps_remote, $system->{vpsid});
	    }
	    @sorted_includes = sort { $a <=> $b } @vps_remote;
	}

	# Handle early exit if nothing to do:
	if ($noremote == "1") {
	    print STDERR "No KVM has remote backup enabled.\n";
	    print EMAIL "No KVM has remote backup enabled. \n\n";
	}
	else {
	    # Start the backup of KVM VPS's:
	    foreach $vpsid (@sorted_includes) {
		$remote_target_dir_vps = $remote_location_rsync . "/" . $vpsid;
        	($sys_oid) = $cce->find('Vserver', {'vpsid' => $vpsid});
        	($ok, $sysvserver) = $cce->get($sys_oid);
		print STDERR "Performing remote backup of KVM $vpsid to $remote_target_dir_vps\n";

		# Handle --bwlimit:
		if (($solbackupValues->{'RsyncSpeedLimit'} ne "0") && ($solbackupValues->{'RsyncSpeedLimit'} ne "")) {
		    $RsyncSpeedLimit = $solbackupValues->{'RsyncSpeedLimit'};
		    $bwlimit = "--bwlimit=$RsyncSpeedLimit";
		}
		else {
		    $bwlimit = "";
		}

		# FQDN or IP:
    		if ($targethost) {
        	    $target = $targethost;
    		}
    		else {
        	    $target = $targetip;
    		}
		print EMAIL "Performing remote backup of KVM $vpsid to $target:$remote_target_dir_vps\n";

		# Get KVM settings:
		@kvmoid = $cce->find('KVM', {'vpsname' => $vpsid});
		($ok, $kvm) = $cce->get($kvmoid[0]);
		$kvmdisk = $kvm->{'disk'};

		# Get just the name of the filename of the disk image:
		$base_name = basename($kvmdisk);

		if (($solbackupValues->{'vzdump_available'} == "1") && ($solbackupValues->{'vzdump_enabled'} == "1")) {

		    # Back up via Net::SFTP::Foreign:
		
    		    # Connect to remote storage:
    		    no warnings 'Net::SFTP::Foreign';
    		    $sftp = Net::SFTP::Foreign->new($target, password => $pass, port => $port, autodie => 0, stderr_fh => $ssherr);
    		    $sftp->die_on_error("Unable to connect to remote host");

        	    # Does the target directory $remote_target_dir_vps exist on the remote storage?
        	    $fh1 = $sftp->stat("$remote_target_dir_vps");
        	    if ($fh1) {
			# Directory $remote_target_dir_vps exists
		    }
		    else {
			# Creating missing $remote_target_dir_vps on storage:
			$fh2 = $sftp->mkpath("$remote_target_dir_vps");
    		    }

    		    # Transfer KVM Disk image to remote storage:
    		    $remote_file = $remote_target_dir_vps . "/" . $base_name;
    		    $fh1 = $sftp->put($kvmdisk, $remote_file, late_set_perm => 1);
    		    # or die "Unable to transfer $kvmdisk to remote storage: " . $sftp->error;

    		    # Transfer KVM XML config to remote storage:
    		    $remote_xml = $remote_target_dir_vps . "/$vpsid.xml";
    		    $fh1 = $sftp->put("/etc/libvirt/qemu/$vpsid.xml", $remote_xml, late_set_perm => 1);
    		    # or die "Unable to transfer "/etc/libvirt/qemu/$vpsid.xml" to remote storage: " . $sftp->error;
		}
		else {
		    # Back up via RSYNC:
		    system("/usr/bin/ssh -p $port $target mkdir -p $remote_target_dir_vps> /dev/null 2>&1");
		    system("/usr/bin/rsync -q -ar $bwlimit -e 'ssh -p $port' $kvmdisk $target:$remote_target_dir_vps/$base_name --delete-after");
		    system("/usr/bin/rsync -q -ar $bwlimit -e 'ssh -p $port' /etc/libvirt/qemu/$vpsid.xml $target:$remote_target_dir_vps/$vpsid.xml --delete-after");
		}

		# As remote backup has just ended, we now update CODB with the info that says so:
        	@vps_oid = $cce->find('Backup', {'vpsid' => $vpsid, 'origin' => 'rsync', 'location' => 'remote', 'virttype' => 'kvm'});
        	if ($#vps_oid < 0) {
		    # No CODB object for this type of backup yet. Creating one:
		    ($ok) = $cce->create('Backup', {
    			'origin' => 'rsync',
    			'vpsid' => $vpsid,
    			'virttype' => 'kvm',
    			'location' => 'remote',
    			'LBtimestamp' => $datetime,
    			'storage' => $remote_target_dir_vps,
    			'ip' => '',
    			'hostname' => '',
    			'timestamp' => time()
		    });
        	}
		else {
		    # Object already present - updating it:
		    ($ok) = $cce->set($vps_oid[0], '', {
    			'vpsid' => $vpsid,
    			'LBtimestamp' => $datetime,
    			'storage' => $remote_target_dir_vps,
    			'ip' => '',
    			'hostname' => '',
    			'timestamp' => time()
		    });
		}
		print STDERR "Finished remote backup of KVM $vpsid \n\n";
		print EMAIL "Finished remote backup of KVM $vpsid \n\n";
	    }
	}
    }
}

## New VZdump type backup:
sub type_vzdump {

    # Run the separate RSYNC based KVM backup first:
    &backup_kvm_mailhead;
    &backup_kvm_local;
    &backup_kvm_remote;
    &backup_kvm_mailfoot;

    # Now run the native VZdump backup:

    # Define values:
    $vzdump_available		= $solbackupValues->{'vzdump_available'};
    $backup_part_available	= $solbackupValues->{'backup_part_available'};
    $vzdump_enabled		= $solbackupValues->{'vzdump_enabled'};
    $vzdump_stdexcludes		= $solbackupValues->{'vzdump_stdexcludes'};
    $vzdump_backup_type		= $solbackupValues->{'vzdump_backup_type'};
    $vzdump_maxfiles		= $solbackupValues->{'vzdump_maxfiles'};
    $vzdump_ionice		= $solbackupValues->{'vzdump_ionice'};
    $vzdump_bwlimit		= $solbackupValues->{'vzdump_bwlimit'};
    $vzdump_lockwait		= $solbackupValues->{'vzdump_lockwait'};
    $vzdump_stopwait		= $solbackupValues->{'vzdump_stopwait'};
    $backupdir			= $solbackupValues->{'MN_BackupLocalLocation'} . "/vzbackup";
    $backupdir			=~ s|//|/|g;
    $dumpdir			= $solbackupValues->{'MN_BackupLocalLocation'} . "/vzdump";
    $dumpdir			=~ s|//|/|g;

    # Make sure the needed subdirectories are present:
    if (!-d $backupdir) {
	system("mkdir -p $backupdir");
    }
    if (!-d $dumpdir) {
	system("mkdir -p $dumpdir");
    }

    # Can we do snapshots?
    if ($vzdump_backup_type == "1") {
	$dumptype = "--snapshot";
    }
    else {
	$dumptype = "--suspend";
    }

    # Excludes?
    if ($vzdump_stdexcludes == "1") {
	$sw_excludes = " --stdexcludes";
    }

    # Find all VPS's where local backup is enabled:
    $nolocal = "0";
    @oids = $cce->find('Vserver_Backup', {'VserverBackupLocalEnabled' => "1"});
    if ($#oids == -1) {
        $nolocal = "1";
    }
    else {
	# Get all VPS's with enabled local backup:
	foreach $line (@oids) {
    	    ($ok, $system) = $cce->get($line);
    	    $func = push(@vps_local, $system->{vpsid});
	}
    }

    # Find all VPS's where remote  backup is enabled:
    @oids = $cce->find('Vserver_Backup', {'VserverBackupRemoteEnabled' => "1"});
    if ($#oids == -1) {
        $noremote = "1";
    }
    else {
        # Get all VPS's with enabled remote backup:
        foreach $line (@oids) {
	    ($ok, $system) = $cce->get($line);
    	    $func = push(@vps_remote, $system->{vpsid});
	}
    }

    # Handle early exit if nothing to do:
    if (($nolocal == "1") && ($noremote == "1")) {
	print STDERR "No VPS has local or remote backup enabled. Nothing to do!\n";
	system("/bin/rm -f /tmp/vz_vserver_backup.lock");
	exit;
    }

    # Merge arrays and remove duplicates:
    @joblist = uniq(@vps_local, @vps_remote);

    # Find all VPS's and list them in an array:
    @oids = $cce->find('Vserver');
    foreach $line (@oids) {
        ($ok, $system) = $cce->get($line);
        $func = push(@full_list, $system->{vpsid});
    }
    @oids = $cce->find('KVM');
    foreach $line (@oids) {
        ($ok, $system) = $cce->get($line);
        $func = push(@full_list, $system->{vpsid});
    }
	
    # Now build a list of all VPS's that we need to exclude from the backup creation:
    @isect 	= ( );
    @excludes  	= ( );
    %count	= ( );
    foreach $item (@joblist, @full_list) { $count{$item}++;}
    foreach $item (keys %count) {
	if ($count{$item} == 2) {
    	    push @isect, $item;
	} 
	else {
    	    push @excludes, $item;
	}
    }
    @sorted_excludes = sort { $a <=> $b } @excludes;
    @sorted_includes = sort { $a <=> $b } @isect;

    if ($DEBUG == "1") {
	print "sorted_excludes " . join(" ", @sorted_excludes), "\n";
	print "elements: " . $#excludes . "\n";
	print "sorted_includes " . join(" ", @sorted_includes), "\n";
    }

    # Define what we back up (--all or a list of selected VPS's):
    if ($#sorted_excludes == -1) {
	$what = "--all";
    }
    elsif ($#sorted_includes > -1) {
	$what = join(" ", @sorted_includes);
    }
    else {
	print STDERR "Nothing to do!\n";
	system("/bin/rm -f /tmp/vz_vserver_backup.lock");
	exit;
    }

	# Define hook-script:
	$hook = "--script /usr/sausalito/sbin/vzbackup-vzdump-hook.pl";

    if (!$vzdump_bwlimit) { $vzdump_bwlimit = "4096"; }

    # Do the actual backup:
    system("/usr/sbin/vzdump --dumpdir $backupdir --maxfiles $vzdump_maxfiles --tmpdir $dumpdir --ionice $vzdump_ionice --bwlimit $vzdump_bwlimit --lockwait $vzdump_lockwait --stopwait $vzdump_stopwait $sw_excludes --mailto admin $dumptype $what $hook");

    # VZdump housecleaning:
    &clean_up_vzdumps;
}

## New RSYNC type backup below:

sub type_rsync {

    # Email generation:
    open (EMAIL, '>>/tmp/.backup_mail');
    print EMAIL "=================================================================\n";
    print EMAIL "Aventurin{e} $B_type Backup on $host ($addr) \n";
    print EMAIL "Starting up at: $now\n";
    print EMAIL "=================================================================\n\n";

    # Get settings for RSYNC related backup:
    &rsync_settings;

    # Handle local RSYNC backup if local backup is enabled and supposed to run now:
    if (($MN_LocalBackupEnabled == "1") && (($s_type eq "local") || ($s_type eq "both"))) {

	if ($kvm_support eq "1") {
	    &backup_kvm_local;
	}

	print STDERR "Starting local RSYNC backup of VPS's ... \n\n";
	print EMAIL "Starting local RSYNC backup of VPS's ... \n";
	print EMAIL "=========================================\n\n";

	# Find all OpenVZ VPS's where local backup is enabled:
	$nolocal = "0";
	@oids = $cce->find('Vserver_Backup', {'VserverBackupLocalEnabled' => "1", 'virttype' => 'openvz'});
	if ($#oids == -1) {
    	    $nolocal = "1";
	}
	else {
	    # Get all VPS's with enabled local backup:
	    undef @vps_local;
	    undef @sorted_includes;
	    foreach $line (@oids) {
    		($ok, $system) = $cce->get($line);
    		$func = push(@vps_local, $system->{vpsid});
	    }
	    @sorted_includes = sort { $a <=> $b } @vps_local;
	}

	# Handle early exit if nothing to do:
	if ($nolocal == "1") {
	    print STDERR "No VPS has local backup enabled. \n";
	    print EMAIL "No VPS has local backup enabled. \n\n";
	}
	else {

	    # Start the backup of OpenVZ containers:
	    foreach $vpsid (@sorted_includes) {
		$local_target_dir_vps = $local_location_rsync . "/" . $vpsid;
        	($sys_oid) = $cce->find('Vserver', {'vpsid' => $vpsid});
        	($ok, $sysvserver) = $cce->get($sys_oid);
		$vps_fqdn = $sysvserver->{'hostname'};
		$vps_ip = $sysvserver->{'ip'};
		print STDERR "Performing local backup of VPS $vpsid to $local_target_dir_vps\n";
		print EMAIL "Performing local backup of VPS $vpsid to $local_target_dir_vps\n";
		if (!-d "$local_target_dir_vps") {	
		    system("mkdir -p $local_target_dir_vps");
		}
		system("/usr/bin/rsync -q -ar /vz/private/$vpsid $local_location_rsync --delete-after");
		if (!-d "$local_target_dir_vps/etc/vzdump") {
		    system("mkdir -p $local_target_dir_vps/etc/vzdump");
		}
		if (!-e "$local_target_dir_vps/etc/vzdump/vps.conf") {
		    system("cp /etc/vz/conf/$vpsid.conf $local_target_dir_vps/etc/vzdump/vps.conf");
		}

		# As local backup has just ended, we now update CODB with the info that says so:
        	@vps_oid = $cce->find('Backup', {'vpsid' => $vpsid, 'origin' => 'rsync', 'location' => 'local', 'virttype' => 'openvz'});
        	if ($#vps_oid < 0) {
		    # No CODB object for this type of backup yet. Creating one:
		    ($ok) = $cce->create('Backup', {
    			'origin' => 'rsync',
    			'vpsid' => $vpsid,
    			'virttype' => 'openvz',
    			'location' => 'local',
    			'LBtimestamp' => $datetime,
    			'storage' => $local_target_dir_vps,
    			'ip' => $vps_ip,
    			'hostname' => $vps_fqdn,
    			'timestamp' => time()
		    });
        	}
		else {
		    # Object already present - updating it:
		    ($ok) = $cce->set($vps_oid[0], '', {
    			'vpsid' => $vpsid,
    			'LBtimestamp' => $datetime,
    			'storage' => $local_target_dir_vps,
    			'ip' => $vps_ip,
    			'hostname' => $vps_fqdn,
    			'timestamp' => time()
		    });
		}
		print STDERR "Finished local backup of VPS $vpsid \n\n";
		print EMAIL "Finished local backup of VPS $vpsid \n\n";
	    }
	}
    }

    # Handle remote RSYNC backup if remote backup is enabled and supposed to run now:
    if (($MN_RemoteBackupEnabled == "1") && (($s_type eq "remote") || ($s_type eq "both"))) {

	if ($kvm_support eq "1") {
	    &backup_kvm_remote;
	}

	print STDERR "Starting remote RSYNC backup of VPS's ... \n\n";
	print EMAIL "Starting remote RSYNC backup of VPS's ... \n";
	print EMAIL "=========================================\n\n";

	# Find all OpenVZ VPS's where remote backup is enabled:
        $noremote = "0";
	@oids = $cce->find('Vserver_Backup', {'VserverBackupRemoteEnabled' => "1", 'virttype' => 'openvz'});
	if ($#oids == -1) {
    	    $noremote = "1";
	}
	else {
	    # Get all VPS's with enabled remote backup:
	    undef @vps_remote;
	    undef @sorted_includes;
	    foreach $line (@oids) {
    		($ok, $system) = $cce->get($line);
                $vps_config = "/etc/vz/conf/" . $system->{vpsid} . ".conf";
                if (-f $vps_config) {
			# VPS exists, push it into array for backing up:
                        $func = push(@vps_remote, $system->{vpsid});
                }
		else {
			# VPS does not exists. This must be a backup entry from a removed VPS.
			# Remove the 'Vserver_Backup' entry as it is no longer valid:
			($ok) = $cce->destroy($line);
		}
	    }
	    @sorted_includes = sort { $a <=> $b } @vps_remote;
	}

	# Handle early exit if nothing to do:
	if ($noremote == "1") {
	    print STDERR "No VPS has remote backup enabled.\n";
	    print EMAIL "No VPS has remote backup enabled. \n\n";
	}
	else {
	    # Start the backup of OpenVZ containers:
	    foreach $vpsid (@sorted_includes) {
		$remote_target_dir_vps = $remote_location_rsync . "/" . $vpsid;
        	($sys_oid) = $cce->find('Vserver', {'vpsid' => $vpsid});
        	($ok, $sysvserver) = $cce->get($sys_oid);
		$vps_fqdn = $sysvserver->{'hostname'};
		$vps_ip = $sysvserver->{'ip'};
		print STDERR "Performing remote backup of VPS $vpsid to $remote_target_dir_vps\n";

		# Copy VPS config into container:
		if (!-d "/vz/private/$vpsid/etc/vzdump") {
		    system("mkdir -p /vz/private/$vpsid/etc/vzdump");
		}
		if (!-e "/vz/private/$vpsid/etc/vzdump/vps.conf") {
		    system("cp /etc/vz/conf/$vpsid.conf /vz/private/$vpsid/etc/vzdump/vps.conf");
		}

		# Handle --bwlimit:
		if (($solbackupValues->{'RsyncSpeedLimit'} ne "0") && ($solbackupValues->{'RsyncSpeedLimit'} ne "")) {
		    $RsyncSpeedLimit = $solbackupValues->{'RsyncSpeedLimit'};
		    $bwlimit = "--bwlimit=$RsyncSpeedLimit";
		}
		else {
		    $bwlimit = "";
		}

		# FQDN or IP:
    		if ($targethost) {
        	    $target = $targethost;
    		}
    		else {
        	    $target = $targetip;
    		}
		print EMAIL "Performing remote backup of VPS $vpsid to $target:$remote_target_dir_vps\n";

		# Back up VPS:
		system("/usr/bin/rsync -q -ar $bwlimit -e 'ssh -p $port' /vz/private/$vpsid $target:$remote_location_rsync/ --delete-after");

		# As remote backup has just ended, we now update CODB with the info that says so:
        	@vps_oid = $cce->find('Backup', {'vpsid' => $vpsid, 'origin' => 'rsync', 'location' => 'remote', 'virttype' => 'openvz'});
        	if ($#vps_oid < 0) {
		    # No CODB object for this type of backup yet. Creating one:
		    ($ok) = $cce->create('Backup', {
    			'origin' => 'rsync',
    			'vpsid' => $vpsid,
    			'virttype' => 'openvz',
    			'location' => 'remote',
    			'LBtimestamp' => $datetime,
    			'storage' => $remote_target_dir_vps,
    			'ip' => $vps_ip,
    			'hostname' => $vps_fqdn,
    			'timestamp' => time()
		    });
        	}
		else {
		    # Object already present - updating it:
		    ($ok) = $cce->set($vps_oid[0], '', {
    			'vpsid' => $vpsid,
    			'LBtimestamp' => $datetime,
    			'storage' => $remote_target_dir_vps,
    			'ip' => $vps_ip,
    			'hostname' => $vps_fqdn,
    			'timestamp' => time()
		    });
		}
		print STDERR "Finished remote backup of VPS $vpsid \n\n";
		print EMAIL "Finished remote backup of VPS $vpsid \n\n";
	    }
	}
    }

    $now = localtime time;
    print EMAIL "=================================================================\n";
    print EMAIL "Aventurin{e} $B_type Backup on $host ($addr) \n";
    print EMAIL "Backup finished at: $now\n";
    print EMAIL "=================================================================\n\n";
    close (EMAIL);

    # Run extender:
    if ( -f "/usr/sausalito/sbin/vz_vserver_backup_extend.pl") {
	system("/usr/sausalito/sbin/vz_vserver_backup_extend.pl");
    }

    # Send the email:
    open (EMAIL, '/tmp/.backup_mail');
	while (<EMAIL>) {
	    $body .= $_;
	}
    close (EMAIL);
    system("/bin/rm -f /tmp/.backup_mail");

    if ($body) {
	$body = $body_head . $body;
	$subject = $host . ": rsync backup finished";

	# Build the message using MIME::Lite instead:
	my $send_msg = MIME::Lite->new(
    	    From     => "root <root>",
    	    To       => "root <root>",
    	    Subject  => $subject,
    	    Data     => $body
	);   
        
	# Set content type:
	$send_msg->attr("content-type"         => "text/plain");
	$send_msg->attr("content-type.charset" => "ISO-8859-1"); 
     
	# Out with the email:
	$send_msg->send;
    }
}

sub clean_up_vzdumps {

    # Define values:
    $vzdump_enabled             = $solbackupValues->{'vzdump_enabled'};
    $rsync_enabled              = $solbackupValues->{'rsync_enabled'};
    $MN_RemoteBackupEnabled     = $solbackupValues->{'MN_RemoteBackupEnabled'};
    $MN_LocalBackupEnabled      = $solbackupValues->{'MN_LocalBackupEnabled'};
    $MN_BackupLocalLocation     = $solbackupValues->{'MN_BackupLocalLocation'};
    $targethost                 = $solbackupValues->{'MN_BackupRemoteTargetHost'};
    $targetip                   = $solbackupValues->{'MN_BackupRemoteTargetIP'};
    $path                       = $solbackupValues->{'MN_BackupRemoteLocation'};
    $user                       = $solbackupValues->{'MN_BackupRemoteTargetUser'};
    $pass                       = $solbackupValues->{'MN_BackupRemoteTargetUserPass'};
    $port                       = $solbackupValues->{'MN_BackupRemoteTargetPort'};
    $limit                      = $solbackupValues->{'RsyncSpeedLimit'};
    $my_maxfiles                = $solbackupValues->{'vzdump_maxfiles'};

    # Assemble paths:  
    $local_location_vzdump = $solbackupValues->{'MN_BackupLocalLocation'} . "/vzbackup";
    $local_location_vzdump =~ s|//|/|g;
    $local_location_vzdump =~ s|/$||g;

    $remote_location_vzdump = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_vzdump =~ s|//|/|g;
    $remote_location_vzdump =~ s|/$||g;

    $local_location_rsync = $solbackupValues->{'MN_BackupLocalLocation'} . "/rsync";
    $local_location_rsync =~ s|//|/|g;  
    $local_location_rsync =~ s|/$||g;

    $remote_location_rsync = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_rsync =~ s|//|/|g;
    $remote_location_rsync =~ s|/$||g;

    # Remove remote CODB 'Backup' tarballs if we have more 'remote' Backup objects than 
    # --maxfiles allows. First build an array with all 'remote' 'Backup' objects of this
    # VPSID:

    @vpsids_remote_backups_all = $cce->findx('Backup', {'location' => 'remote', 'origin' => 'vzdump'});
    for $finding (@vpsids_remote_backups_all) {
	($ok, $back_obj) = $cce->get($finding);
	$vmid = $back_obj->{'vpsid'};
	push(@all_vpsids, $vmid);
    }

    @my_vpsids = uniq(@all_vpsids);

    for $vmid (@my_vpsids) {
	$num_of_remotes = "0";
	@my_remote_backups = $cce->findx('Backup', {'vpsid' => $vmid, 'location' => 'remote', 'origin' => 'vzdump'});
	for $entry (@my_remote_backups) {
    	    $num_of_remotes++;
	}

	# Remove excess remote backups:
	@sorted = ();
	@remote_stored = ();
	if ($num_of_remotes > $my_maxfiles) {
    	    # We have more remote backups than we want to keep:
    	    for $xentry (@my_remote_backups) {
                ($ok, $old_entry) = $cce->get($xentry);
                push(@remote_stored, $old_entry->{'timestamp'});
    	    }
    	    @sorted = sort { $a <=> $b } @remote_stored;

    	    # Remove as many newest entries from @sorted as we have maxfiles set to:
    	    $z = $my_maxfiles;
    	    for ($z = $my_maxfiles; $z > 0; $z--) {
        	pop(@sorted);
    	    }

    	    # We now have a sorted array with the remote backups. Remove the oldest tarball(s):
    	    foreach $goner (@sorted) {
        	@my_single = $cce->find('Backup', {'vpsid' => $vmid, 'timestamp' => $goner, 'location' => 'remote', 'origin' => 'vzdump'});
        	($ok, $expire_goner) = $cce->get($my_single[0]);
        	$exp_path = $expire_goner->{'storage'};

        	# Assemble target login details:
        	if ($targethost) {
            	    $target = $user . '@' . $targethost;
        	}
        	else {
            	    $target = $user . '@' . $targetip;
        	}

        	# Connect to remote storage:
        	no warnings 'Net::SFTP::Foreign';
        	$sftp = Net::SFTP::Foreign->new($target, password => $pass, port => $port, autodie => 0, stderr_fh => $ssherr);
        	$sftp->die_on_error("Unable to connect to remote host");

        	# Does the old file exist?
        	$fh1 = $sftp->stat("$exp_path");
        	if ($fh1) {
            	    # Remove file:
            	    $fh2 = $sftp->remove("$exp_path");
            	    if ($fh2) {
                	# Remove CODB Object:
                	($ok) = $cce->destroy($my_single[0]);
            	    }
        	}
    	    }
	}
    }
}

# General Settings for RSYNC related backup:
sub rsync_settings {

    # Define values:
    $vzdump_enabled             = $solbackupValues->{'vzdump_enabled'};
    $rsync_enabled              = $solbackupValues->{'rsync_enabled'};
    $MN_RemoteBackupEnabled     = $solbackupValues->{'MN_RemoteBackupEnabled'};
    $MN_LocalBackupEnabled      = $solbackupValues->{'MN_LocalBackupEnabled'};
    $MN_BackupLocalLocation     = $solbackupValues->{'MN_BackupLocalLocation'};
    $targethost                 = $solbackupValues->{'MN_BackupRemoteTargetHost'};
    $targetip                   = $solbackupValues->{'MN_BackupRemoteTargetIP'};
    $path                       = $solbackupValues->{'MN_BackupRemoteLocation'};
    $user                       = $solbackupValues->{'MN_BackupRemoteTargetUser'};
    $pass                       = $solbackupValues->{'MN_BackupRemoteTargetUserPass'};
    $port                       = $solbackupValues->{'MN_BackupRemoteTargetPort'};
    $limit                      = $solbackupValues->{'RsyncSpeedLimit'};
    $my_maxfiles                = $solbackupValues->{'vzdump_maxfiles'};
    $kvm_support                = $solbackupValues->{'kvm_support'};

    # Assemble paths:  
    $local_location_vzdump = $solbackupValues->{'MN_BackupLocalLocation'} . "/vzbackup";
    $local_location_vzdump =~ s|//|/|g;
    $local_location_vzdump =~ s|/$||g;

    $remote_location_vzdump = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_vzdump =~ s|//|/|g;
    $remote_location_vzdump =~ s|/$||g;

    $local_location_rsync = $solbackupValues->{'MN_BackupLocalLocation'} . "/rsync";
    $local_location_rsync =~ s|//|/|g;  
    $local_location_rsync =~ s|/$||g;

    $remote_location_rsync = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_rsync =~ s|//|/|g;
    $remote_location_rsync =~ s|/$||g;

    # Local directory setup:
    if (!-d $local_location_rsync) {
	system("mkdir -p $local_location_rsync");
    }
    if (!-w $local_location_rsync) {
	print STDERR "Specified backup location $solbackupValues->{'MN_BackupLocalLocation'} is not writeable!\n";
	print STDERR "Defaulting backup location to /vz/backup \n\n";
	$local_location_rsync = "/vz/backup/rsync";
    }
}

$cce->bye('SUCCESS');
exit 0;
